@@ -1256,7 +1256,7 @@ void irc_msg_ping(struct irc_conn *irc, const char *name, const char *from, char
 	char *buf;
 
 	buf = irc_format(irc, "v:", "PONG", args[0]);
-	irc_send(irc, buf);
+	irc_priority_send(irc, buf);
 	g_free(buf);
 }
 
@@ -1454,10 +1454,20 @@ irc_sasl_cb_simple(void *ctx, int id, const char **res, unsigned *len)
 {
 	struct irc_conn *irc = ctx;
 	PurpleConnection *gc = purple_account_get_connection(irc->account);
+	const char *saslname =
+		purple_account_get_string(irc->account, "saslname", NULL);
+
+	if(saslname == NULL || *saslname == '\0') {
+		saslname = purple_account_get_string(irc->account, "realname", "");
+	}
+
+	if(saslname == NULL || *saslname == '\0') {
+		saslname = purple_connection_get_display_name(gc);
+	}
 
 	switch(id) {
 		case SASL_CB_AUTHNAME:
-			*res = purple_connection_get_display_name(gc);
+			*res = saslname;
 			break;
 		case SASL_CB_USER:
 			*res = "";
@@ -1569,7 +1579,7 @@ irc_auth_start_cyrus(struct irc_conn *irc)
 	purple_debug_info("irc", "Using SASL: %s\n", irc->current_mech);
 
 	buf = irc_format(irc, "vv", "AUTHENTICATE", irc->current_mech);
-	irc_send(irc, buf);
+	irc_priority_send(irc, buf);
 	g_free(buf);
 }
 
@@ -1700,7 +1710,7 @@ irc_msg_auth(struct irc_conn *irc, char *arg)
 		authinfo = g_strdup("+");
 
 	buf = irc_format(irc, "vv", "AUTHENTICATE", authinfo);
-	irc_send(irc, buf);
+	irc_priority_send(irc, buf);
 	g_free(buf);
 	g_free(authinfo);
 	g_free(serverin);
@@ -1723,7 +1733,7 @@ irc_msg_authok(struct irc_conn *irc, const char *name, const char *from, char **
 
 	/* Finish auth session */
 	buf = irc_format(irc, "vv", "CAP", "END");
-	irc_send(irc, buf);
+	irc_priority_send(irc, buf);
 	g_free(buf);
 }
 
@@ -1802,7 +1812,7 @@ irc_sasl_finish(struct irc_conn *irc)
 
 	/* Auth failed, abort */
 	buf = irc_format(irc, "vv", "CAP", "END");
-	irc_send(irc, buf);
+	irc_priority_send(irc, buf);
 	g_free(buf);
 }
 #endif
